/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <sys/stat.h>
#ifdef	MALLOC_H
#include <malloc.h>
#endif
#include "is_types.h"
#include "isam.h"

#ifdef BSD
#define memcpy(a, b, c) bcopy((b), (a), (c))
#include <strings.h>
#else
#include <memory.h>
#include <string.h>
#endif /* BSD */

#ifndef lint
static	char	sccsid[] = "@(#)is_add.c	1.10	10:32:50	2/26/92";
#endif

#ifndef	MALLOC_H
extern	char	*malloc();
extern	char	*realloc();
#endif

extern	int	db_errno;

void
__isam_key_insert (keyblk, key, i, data, record)
keyblk_t *keyblk;
sort_key_t *key;
int	i;
VOIDP	*data;
off_t	record;
{
	int	size = key->k_key_size + sizeof record;
	register char	*new, *cp, *end;

	new = keyblk->kb_data + i * size;

	if (keyblk->kb_cnt != 0 && keyblk->kb_cnt != i) {
		end = keyblk->kb_data + (keyblk->kb_cnt + 1) * size;
		cp = end - size;

		while (cp != new)
			*--end = *--cp;
	}
	memcpy ((void *) new, (void *) &record, sizeof record);
	memcpy ((void *) (new + sizeof record), data, key->k_key_size);

	keyblk->kb_cnt++;
}

void
__isam_key_delete (keyblk, key, i, cnt)
keyblk_t *keyblk;
sort_key_t *key;
int	i;
int	cnt;
{
	int	size = key->k_key_size + sizeof (off_t);
	register char	*new, *cp, *end;

	if (i < keyblk->kb_cnt - 1) {
		new = keyblk->kb_data + i * size;
		end = keyblk->kb_data + (keyblk->kb_cnt * size) + 1;

		cp = new + (size * cnt);

		while (cp != end)
			*new++ = *cp++;
	}
	keyblk->kb_cnt -= cnt;
}

/*
 * __isam_add_eof - add key to end of index
 *
 *	__isam_add_eof must re-traverse the index because the lookup
 *	code doesn't say where the key goes when a match isn't found.
 */

int
__isam_add_eof (file, key, node, data, record)
file_t	*file;
sort_key_t *key;
off_t	node;
VOIDP	*data;
off_t	record;
{
	struct	stat	stat;
	keyblk_t buf;
	keyblk_t *keyblk = &buf;
	off_t	prev_node;
	int	size = key->k_key_size + sizeof record;
	int	cnt;
	int	rc;

	/*
	 * Don't have valid offsets for the blocks because the key
	 * wasn't found.  I have to traverse the tree again (big loss)
	 * to find the end of the index.
	 */

	if (fstat (file->f_key_fd, &stat))
		return (db_errno = DB_EIO), -1;

	if (stat.st_size < (file->f_key_index + 1) * file->f_bsize) {

		/*
		 * I need to add the root node for this index
		 */

		keyblk->kb_cnt = 0;
		keyblk->kb_prev = keyblk->kb_left = keyblk->kb_right = 0;
		return __isam_add (file, key, node, data, record, 0);
	} else {
		
		/*
		 * Traverse the tree to find the lowest-most right-most
		 * key block.  That's the block I want to add this key
		 * to.
		 */

		while (node > 0) {
			if (isam_read (file, node, keyblk))
				return -1;

			prev_node = node;

			if (keyblk->kb_cnt == 0) {
				if (keyblk->kb_prev != 0)
					return (db_errno = DB_BAD_DB), -1;
				else
					break;
			}
			memcpy ((void *) &node, (void *) (keyblk->kb_data +
				(keyblk->kb_cnt - 1) * size), sizeof node);
		}
	}
	return __isam_add (file, key, prev_node, data, record, keyblk->kb_cnt);
}

int
__isam_fix_child (file, key, parent, node)
file_t	*file;
sort_key_t *key;
keyblk_t *parent;
off_t	node;
{
	int	i;
	int	size = key->k_key_size + sizeof (off_t);
	keyblk_t child;
	off_t	record;

	for (i = 0;i < parent->kb_cnt;i++) {
		memcpy ((char *) &record, parent->kb_data + i * size,
			sizeof (off_t));

		if (record < 0)
			continue;

		if (isam_read (file, record, &child))
			return -1;

		if (child.kb_prev == node)
			continue;

		child.kb_prev = node;
		if (isam_write (file, record, &child))
			return -1;
	}
	return 0;
}

int
__isam_fix (file, key, child, record, parent)
file_t	*file;
sort_key_t *key;
keyblk_t *child;
off_t	record;
keyblk_t *parent;
{
	int	i;
	int	rc;
	int	size = key->k_key_size + sizeof (off_t);
	off_t	tmp;
	keyblk_t new, *buf;

	if (parent == 0)
		buf = &new;
	else
		buf = parent;

	if (isam_read (file, child->kb_prev, buf))
		return -1;

	for (i = 0;i < buf->kb_cnt;i++) {
		memcpy ((char *) &tmp, buf->kb_data + size * i, sizeof (off_t));
		if (tmp == record) {
			memcpy (buf->kb_data + size * i + sizeof (off_t),
				child->kb_data + sizeof (off_t),
				key->k_key_size);
			break;
		}
	}
	if (i == buf->kb_cnt)
		return -1;

	if (isam_write (file, child->kb_prev, buf))
		return -1;

	if (i == 0 && buf->kb_prev > 0) {
		if (__isam_fix (file, key, buf,
				child->kb_prev, (keyblk_t *) 0) < 0) {
			return -1;
		}
	}
	return buf->kb_cnt;
}

int
__isam_add_split (file, key, node, data, record, where)
file_t	*file;				/* isam file being updated */
sort_key_t *key;			/* key description */
off_t	node;				/* isam node number being split */
VOIDP	*data;				/* pointer to new key data */
off_t	record;				/* new key record number */
int	where;				/* offset in keyblk to insert key */
{
	keyblk_t *keyblk, buffer;
	keyblk_t new, root;
	int	size = key->k_key_size + sizeof (off_t);
	off_t	new_offset;
	off_t	parent;
	int	new_number;
	int	i, j;
	int	max = (file->f_bsize - (sizeof (long) * 3 + sizeof (int))) /
			size;
	int	rc;

	/*
	 * Get the block being changed
	 */

	keyblk = &buffer;
	if (isam_read (file, node, keyblk))
		return -1;

	/*
	 * Figure out how to split the file - there are three ways.
	 *
	 *	1). BOF - the left pointer is 0 and right > 0.
	 *		Take the first key and make it its own block.  The
	 *		new keys are then updated in the previous block.
	 *	2). EOF - the right pointer is 0 and left > 0.
	 *		Take the last key and make it its own block.  The
	 *		new keys are then updated in the previous block.
	 *	3). OTHER - not BOF or EOF
	 *		Half of the keys are put into one block and
	 *		half into a new block.  If this is the root node,
	 *		it is split and a new root created with just two
	 *		keys.  Otherwise, the previous node is updated to
	 *		reflect the two new children.
	 */

	if (keyblk->kb_left == 0 && keyblk->kb_right > 0) {

		/*
		 * BOF split
		 */

		new.kb_prev = keyblk->kb_prev;
		new.kb_left = 0;
		new.kb_right = node;
		keyblk->kb_left = isam_allocate (file);

		if (where == 0) {
			
			/*
			 * The new key is smaller than the first key.
			 */

			new.kb_cnt = 0;
			__isam_key_insert (&new, key, 0, data, record);
		} else {

			/*
			 * The new key is at least bigger than the
			 * first key.
			 */

			memcpy (new.kb_data, keyblk->kb_data, size);
			new.kb_cnt = 1;

			__isam_key_delete (keyblk, key, 0, 1);
			__isam_key_insert (keyblk, key, where - 1,
				data, record);
		}
		if (isam_write (file, keyblk->kb_left, &new))
			return -1;

		if (__isam_fix_child (file, key, &new, keyblk->kb_left))
			return -1;

		if (__isam_fix_child (file, key, keyblk, node))
			return -1;

		if (isam_write (file, node, keyblk))
			return -1;

		if ((rc = __isam_fix (file, key, keyblk, node, 0)) < 0)
			return -1;

		/*
		 * The current key must be the first key in the parent
		 * node, so there is no need to figure out where to add
		 * this key in the parent.
		 */

		if (__isam_add (file, key, keyblk->kb_prev, new.kb_data +
				sizeof (off_t), - keyblk->kb_left, 0))
			return -1;

		return 0;
	} else if (keyblk->kb_left > 0 && keyblk->kb_right == 0) {

		/*
		 * EOF split
		 */

		new.kb_left = node;
		new.kb_prev = keyblk->kb_prev;
		new.kb_right = 0;
		new.kb_cnt = 0;
		keyblk->kb_right = isam_allocate (file);

		if (where == keyblk->kb_cnt) {

			/*
			 * The new key is bigger than the biggest
			 * key in this block
			 */

			__isam_key_insert (&new, key, 0, data, record);
		} else {

			/*
			 * The new key is smaller than at least the
			 * last key.
			 */

			memcpy (new.kb_data, keyblk->kb_data +
				(keyblk->kb_cnt - 1) * size, size);
			new.kb_cnt = 1;

			keyblk->kb_cnt--;
			__isam_key_insert (keyblk, key, where, data, record);
		}
		if (isam_write (file, keyblk->kb_right, &new))
			return -1;

		if (isam_write (file, node, keyblk))
			return -1;

		if (__isam_fix_child (file, key, &new, keyblk->kb_right))
			return -1;

		if (__isam_fix_child (file, key, keyblk, node))
			return -1;

		/*
		 * Find the location of the current node in its parent.
		 * It is always after the last node in the parent's block.
		 */

		if (isam_read (file, keyblk->kb_prev, &root))
			return -1;

		if (__isam_add (file, key, keyblk->kb_prev,
				new.kb_data + sizeof (off_t),
				- keyblk->kb_right, root.kb_cnt))
			return -1;

		return 0;
	} else {

		/*
		 * MID split
		 */

		new.kb_left = keyblk->kb_left;
		new.kb_right = node;
		new.kb_prev = keyblk->kb_prev;
		keyblk->kb_left = isam_allocate (file);

		if (where < keyblk->kb_cnt / 2) {

			/*
			 * New key is in the first half of the block
			 */

			memcpy (new.kb_data, keyblk->kb_data,
				(keyblk->kb_cnt / 2) * size);
			new.kb_cnt = keyblk->kb_cnt / 2;
			__isam_key_delete (keyblk, key, 0, new.kb_cnt);
			__isam_key_insert (&new, key, where, data, record);
		} else {

			/*
			 * New key is in the second half of the block
			 */

			memcpy (new.kb_data, keyblk->kb_data,
				(keyblk->kb_cnt / 2) * size);
			where -= (new.kb_cnt = keyblk->kb_cnt / 2);
			__isam_key_delete (keyblk, key, 0, new.kb_cnt);
			__isam_key_insert (keyblk, key, where, data, record);
		}

		/*
		 * Now I see if I just split the root node.  Might as
		 * well because all of the needed info is right here.
		 */

		if (new.kb_prev == 0) {
			
			/*
			 * I just split the root node.  Start by
			 * allocating a new node for the right hand
			 * side of the old root node.  Then point
			 * the new children at the parent.
			 */

			root.kb_left = root.kb_right = root.kb_prev = 0;
			new.kb_right = isam_allocate (file);
			keyblk->kb_prev = new.kb_prev = node;

			/*
			 * Hand-craft a new root node.  It will consist
			 * of the first key from each of the two children.
			 * The record numbers will be the blocks for those
			 * two nodes.
			 */

			root.kb_cnt = 0;
			__isam_key_insert (&root, key, 0,
				new.kb_data + sizeof (off_t), keyblk->kb_left);
			__isam_key_insert (&root, key, 1,
				keyblk->kb_data + sizeof (off_t), new.kb_right);

			/*
			 * Now write out the new node parts.  The order
			 * is important.  Write the new node first, that
			 * way if it fails nothing has changed.  Then
			 * write the split part of the old root for the
			 * same reason.  Finally, write the new root node
			 * out.
			 */

			if (isam_write (file, keyblk->kb_left, &new))
				return -1;

			if (isam_write (file, new.kb_right, keyblk))
				return -1;

			if (isam_write (file, node, &root))
				return -1;

			if (__isam_fix_child (file, key, &new, keyblk->kb_left))
				return -1;

			if (__isam_fix_child (file, key, keyblk, new.kb_right))
				return -1;

			return 0;
		}
		if (isam_write (file, keyblk->kb_left, &new))
			return -1;

		if (isam_write (file, node, keyblk))
			return -1;

		if (__isam_fix_child (file, key, &new, keyblk->kb_left))
			return -1;

		if (__isam_fix_child (file, key, keyblk, node))
			return -1;

		/*
		 * Find the location of the current node in its parent.
		 * There is where the new node will be inserted.
		 */

		if (__isam_fix (file, key, keyblk, node, 0) < 0)
			return -1;

		if (isam_read (file, keyblk->kb_prev, &root))
			return -1;

		for (i = 0;i < root.kb_cnt;i++) {
			memcpy (&new_offset, root.kb_data + i * size,
				sizeof (off_t));
			if (new_offset == node)
				break;
		}
		if (__isam_add (file, key, keyblk->kb_prev,
				new.kb_data + sizeof (off_t),
				- keyblk->kb_left, i))
			return -1;

		if (new.kb_left > 0) {
			if (isam_read (file, new.kb_left, &root))
				return -1;

			root.kb_right = keyblk->kb_left;
			if (isam_write (file, new.kb_left, &root))
				return -1;
		}
		return 0;
	}
	/*NOTREACHED*/
}

/* 
 * __isam_add - add a key to an isam block
 *
 *	__isam_add always writes the modified block
 */

int
__isam_add (file, key, node, data, record, where)
file_t	*file;
sort_key_t *key;
off_t	node;
VOIDP	*data;
off_t	record;
int	where;
{
	keyblk_t buf;
	keyblk_t *keyblk = &buf;
	char	*keys;
	int	size = key->k_key_size + sizeof record;
	int	max = (file->f_bsize - (sizeof (long) * 3 + sizeof (int))) /
			size;
	int	cnt;
	int	rc;

	/*
	 * Get the block being added to
	 */

	keyblk = &buf;
	keys = (char *) keyblk->kb_data;

	if (isam_read (file, node, keyblk))
		return -1;

	/*
	 * The the new key won't fit in the current block,
	 * split the block and add the key.
	 */

	if (keyblk->kb_cnt >= max)
		return __isam_add_split (file, key, node,
					data, - record, where);

	/*
	 * Otherwise, insert the key into the current block
	 * and just write it out.
	 */

	__isam_key_insert (keyblk, key, where, data, - record);
	if (isam_write (file, node, keyblk))
		return -1;

	/*
	 * If this is the new first key in that block the parent
	 * will have to be updated.
	 */

	if (where == 0 && keyblk->kb_prev > 0)
		if (__isam_fix (file, key, keyblk, node, 0) < 0)
			return -1;

	if (record < 0) {
		if (__isam_fix_child (file, key, keyblk, node))
			return -1;
	}
	return 0;
}

int
isam_add (file, index, data, record)
file_t	*file;
int	index;
VOIDP	*data;
off_t	record;
{
	int	i;
	sort_key_t *key;
	int	unique;
	int	match;

	/*
	 * Default to the current index if the requested index was -1
	 */

	if (index == -1)
		index = file->f_current_index;

	/*
	 * Find the index to be used.  It must exist, or else
	 * an error is returned.  Block zero (and thus index
	 * zero) is reserved for use as a free index block chain.
	 */

	if (index <= 0 || index > file->f_index->i_key_count)
		return (db_errno = DB_NO_INDEX), -1;

	key = &file->f_index->i_keys[index];
	file->f_key_block = 0;
	file->f_key_slot = 0;

	/*
	 * If the index is unique, the key can't be added if it already
	 * exists in the index.  So, use GREATEREQ for the matching
	 * criteria and see if there was a match for equality.
	 */

	unique = key->k_unique;
	match = unique ? IS_GREATEREQ:IS_GREATER;

	if (__isam_lookup (file, key, index, data,
			match, IS_UNIQUE) == (off_t) -1) {
		return __isam_add_eof (file, key, index, data, record);
	} else {

		/*
		 * See if the lookup found an exact match for a unique
		 * index.  If it did, return an error.
		 */

		if (unique && file->f_key_match == 0)
			return (db_errno = DB_EXISTS), -1;

		return __isam_add (file, key, file->f_key_block,
			data, record, file->f_key_slot);
	}
}

/*
 * isam_add_field - add a new field to an ISAM file
 *
 * isam_add_field() adds a new field definition to the in-core copy of
 * the file's schema.  It does not update the on-disk version.
 */

isam_add_field (file, name, type, len)
file_t	*file;
char	*name;
int	type;
int	len;
{
	field_t	*field;
	schema_t *schema;
	struct	stat	sb;
	int	size;
	int	i;

	switch (type) {
		case T_CHAR:	size = len; break;
		case T_SHORT:	size = sizeof (short); break;
		case T_INT:	size = sizeof (int); break;
		case T_LONG:
		case T_DATE:
		case T_SERIAL:	size = sizeof (long); break;
		case T_FLOAT:	size = sizeof (float); break;
		case T_DOUBLE:
		case T_MONEY:	size = sizeof (double); break;
		default: return (db_errno = DB_INVALID), -1;
	}
	if (type != T_CHAR && len != 0)
		return (db_errno = DB_INVALID), -1;
	else if (type == T_CHAR && (len < 0 || len > 256))
		return (db_errno = DB_INVALID), -1;

	if (file->f_schema == 0) {
		if (! (file->f_schema = (schema_t *)
				malloc (sizeof (schema_t))))
			return (db_errno = DB_NOMEM), -1;

		file->f_schema->s_fields = 0;
		file->f_schema->s_field_count = 0;
	}
	schema = file->f_schema;

	for (i = 0;i < schema->s_field_count;i++) {
		if (strncmp (schema->s_fields[i].f_name,
				name, MAXFIELDNAME) == 0)
			break;
	}
	if (i != schema->s_field_count)
		return (db_errno = DB_FIELD_EXISTS), -1;

	if (schema->s_fields == 0) {
		if (! (schema->s_fields = (field_t *)
				malloc (2 * sizeof (field_t))))
			return (db_errno = DB_NOMEM), -1;

		schema->s_field_count = 2;

		field = &schema->s_fields[1];
	} else {
		if (! (field = (field_t *) realloc (schema->s_fields,
			(schema->s_field_count + 1) * sizeof (field_t))))
			return (db_errno = DB_NOMEM), -1;

		schema->s_fields = field;
		schema->s_field_count++;

		field = &schema->s_fields[i];
	}
	field->f_type = type;
	field->f_size = size;
	field->f_offset = file->f_record_size;
	strncpy (field->f_name, name, MAXFIELDNAME);

	file->f_record_size += size;

	return 0;
}

/*
 * isam_add_index - add a new index to an ISAM file
 */

isam_add_index (file, name, len, fields, unique)
file_t	*file;
char	*name;
int	len;
int	*fields;
int	unique;
{
	field_t	*field;
	schema_t *schema;
	sort_key_t *key;
	is_index_t *index;
	keyblk_t zero;
	struct	stat	sb;
	int	size;
	int	i;

	if ((unique && (len <= 0 || len > MAXSORTFIELDS)) ||
			(! unique && (len < 0 || len >= MAXSORTFIELDS)))
		return (db_errno = DB_TOO_MANY), -1;

	if (file->f_schema == 0) {
		file->f_schema = (schema_t *) malloc (sizeof (schema_t));
		file->f_schema->s_fields = 0;
		file->f_schema->s_field_count = 0;
	}
	schema = file->f_schema;

	for (i = 0;i < schema->s_field_count;i++) {
		if (strncmp (schema->s_fields[i].f_name,
				name, MAXFIELDNAME) == 0)
			break;
	}
	if (i != schema->s_field_count)
		return (db_errno = DB_FIELD_EXISTS), -1;

	for (i = 0;i < len;i++)
		if (fields[i] <= 0 || fields[i] >= schema->s_field_count)
			return (db_errno = DB_INVALID), -1;

	if (file->f_index == 0) {
		if (! (file->f_index = (is_index_t *)
				malloc (sizeof (is_index_t))))
			return (db_errno = DB_NOMEM), -1;

		file->f_index->i_keys = 0;
		file->f_index->i_key_count = 0;
	}
	index = file->f_index;

	if (index->i_keys == 0) {
		if (! (key = (sort_key_t *) malloc (2 * sizeof *key)))
			return (db_errno = DB_NOMEM), -1;

#ifdef BSD
		bzero (key, 2 * sizeof *key);
#else
		memset (key, 0, 2 * sizeof *key);
#endif /* BSD */
		index->i_keys = key;
		index->i_key_count = 1;
	} else {
		if (! (key = (sort_key_t *) realloc (index->i_keys,
				sizeof *key * (index->i_key_count + 2))))
			return (db_errno = DB_NOMEM), -1;

#ifdef BSD
		bzero (key + index->i_key_count + 1, sizeof *key);
#else
		memset (key + index->i_key_count + 1, 0, sizeof *key);
#endif /* BSD */
		index->i_keys = key;
		index->i_key_count++;
	}
	key = index->i_keys + index->i_key_count;

	if (schema->s_fields == 0) {
		if (! (schema->s_fields = (field_t *)
				malloc (2 * sizeof (field_t))))
			return (db_errno = DB_NOMEM), -1;

		schema->s_field_count = 2;

		field = schema->s_fields + 1;
	} else {
		if (! (field = (field_t *) realloc (schema->s_fields,
			(schema->s_field_count + 1) * sizeof (field_t))))
			return (db_errno = DB_NOMEM), -1;

		schema->s_fields = field;
		field = &schema->s_fields[schema->s_field_count++];
	}
	field->f_type = T_INDEX;
	field->f_size = len;
	field->f_offset = index->i_key_count;
	strncpy (field->f_name, name, MAXFIELDNAME);

	key->k_field_number = schema->s_field_count - 1;
	key->k_field_count = len;
	if (! unique)
		key->k_field_count++;

	key->k_unique = unique;
	for (i = 0;i < len;i++)
		key->k_fields[i] = fields[i];

	key->k_key_size = unique ? 0:sizeof (long);
	for (i = 0;i < len;i++)
		key->k_key_size += schema->s_fields[fields[i]].f_size;

	/*
	 * Start the new index off by creating an empty root node.  Take
	 * an index block, zero out the fields, and write it where the
	 * root of that index is.
	 */

#ifdef	BSD
	bzero (&zero, file->f_bsize);
#else
	memset (&zero, 0, file->f_bsize);
#endif
	if (isam_write (file, field->f_offset, &zero))
		return -1;

	return 0;
}
