#!/bin/sh
#
# Test 4
#	Check for unique index functioning
#
#	@(#)test_04.sh	1.6	12:32:11	2/21/92
#
RC=1
set -e

rm -f test_04.sch test_04.idx test_04.dat test.idx test.dat test.out test.in

trap 'rm -f test_04.sch test_04.idx test_04.dat test.idx test.dat test.out test.in test_04 test_04.o ; exit $RC' 0 1 2 15

cc -o test_04 $CCFLAGS test_04.c libidbm.a $LDFLAGS

DBPATH=. ; export DBPATH

cat > test_04.sch << EOF
database test

file test_04

field test_int type int
field test_idx type index unique test_int

end
EOF

#
# create the test schema
#
./dbschema test_04.sch
if [ $? != 0 ]; then
	exit
fi

#
# load some unique records
#
./dbload test test_04 << EOF
1
2
3
4
5
EOF
#
# load some new unique records
#
./dbload test test_04 << EOF
11
12
13
14
15
EOF
#
# load some mixed (unique and non-unique) records
#
cat > test.in << EOF
1
11
21
2
12
22
EOF
./dbload -f test.in test test_04 > test.out 2>&1
#
# check the error messages
#
cmp -s - test.out << EOF
dbload: error writing record 1: IDBM Record exists
dbload: error writing record 2: IDBM Record exists
dbload: error writing record 4: IDBM Record exists
dbload: error writing record 5: IDBM Record exists
EOF
#
# check the database file
#
./dbunload test test_04 test_idx > test.out
cmp -s - test.out << EOF
1
2
3
4
5
11
12
13
14
15
21
22
EOF
#
# delete some of these records, then try to re-add them later
#
./test_04 test test_04 test_idx 2
./test_04 test test_04 test_idx 12
./test_04 test test_04 test_idx 22
#
# check for the correct contents
#
./dbunload test test_04 test_idx > test.out
cmp -s - test.out << EOF
1
3
4
5
11
13
14
15
21
EOF
#
# add five records, including one non-unique one and a completely new one
#
cat > test.in << EOF
1
2
12
22
32
EOF
./dbload -f test.in test test_04 > test.out 2>&1
#
# check for the correct error message
#
cmp -s - test.out << EOF
dbload: error writing record 1: IDBM Record exists
EOF
#
# check for the correct file content
#
./dbunload test test_04 test_idx > test.out
cmp -s - test.out << EOF
1
2
3
4
5
11
12
13
14
15
21
22
32
EOF
RC=0
