#!/bin/sh
#
#	Test 12: test DBUNLOAD error checking
#
#	@(#)test_12.sh	1.1	11:22:05	2/7/92
#
#	This test verifies the error messages produced by DBUNLOAD.
#	Correct function has been verified in other test suites.
#

rm -f test_12.out
RC=1
set -e
trap 'rm -f test_12.out test_12.err ; exit $RC' 0 1 2 3 15

#
# output the titles of the songs.
#
./dbunload -v sng_title music songs > test_12.out
cmp -s - test_12.out << EOF
Early In The Morning
An Innocent Man
Uptown Girl
This Night
Tell Her About It
The Longest Time
Can't Believe Your Lyin'
Christie Lee
Leave a Tender Moment Alone
Keeping The Faith
That's Not Her Style
We Didn't Start The Fire
The Downeaster "Alexa"
I Go To Extremes
Shameless
Storm Front
Leningrad
State Of Grace
When In Rome
And So It Goes
Hang In Long Enough
That's Just The Way It Is
Do You Remember?
Something Happened On The Way To Heaven
Colours
I Wish It Would Rain Down
Another Day In Paradise
Heat On The Street
All Of My Life
Saturday Night And Sunday Morning
Father To Son
Find A Way To My Heart
Sussudio
Only You Know And I Know
Long Long Way To Go
I Don't Wanna Know
One More Night
Don't Lose My Number
Who Said I Would
Doesn't Anybody Stay Together Anymore
Inside Out
Take Me Home
We Said Hello Goodbye
Island Of Souls
All This Time
Mad About You
Why Should I Cry For You
Jeremiah Blues (Part 1)
The Wild Wild Sea
The Soul Cages
Saint Agnes And the Burning Train
When The Angels Fall
Simply Irresistible
More Than Ever
Change His Ways
Coupe De Ville
It Could Happen To You
She Makes My Day
Between Us
Casting A Spell
Tell Me I'm Not Dreaming
Ten Men Workin'
This Note's For You
Life In The City
Twilight
Married Man
Sunny Inside
Hey Hey
One Thing
Disturbing Behavior
Easy Money
Careless Talk
EOF
#
# this time, use the index
#
./dbunload -V sng_pub_key,sng_cat_key,sng_track music songs > test_12.out
cmp -s - test_12.out << EOF
EMI|48057-2|5
CBS|CK38837|2
CBS|CK38837|3
CBS|CK38837|4
CBS|CK38837|5
CBS|CK38837|6
REPRISE|25719-2|8
CBS|CK38837|8
CBS|CK38837|9
CBS|CK38837|10
CBS|CK44366|1
CBS|CK44366|2
CBS|CK44366|3
CBS|CK44366|4
CBS|CK44366|5
CBS|CK44366|6
CBS|CK44366|7
CBS|CK44366|8
CBS|CK44366|9
CBS|CK44366|10
ATLANTIC|82050-2|1
ATLANTIC|82050-2|2
ATLANTIC|82050-2|3
ATLANTIC|82050-2|4
ATLANTIC|82050-2|5
ATLANTIC|82050-2|6
ATLANTIC|82050-2|7
ATLANTIC|82050-2|8
ATLANTIC|82050-2|9
ATLANTIC|82050-2|10
ATLANTIC|82050-2|11
ATLANTIC|82050-2|12
ATLANTIC|81240-2|1
ATLANTIC|81240-2|2
ATLANTIC|81240-2|3
ATLANTIC|81240-2|4
ATLANTIC|81240-2|5
ATLANTIC|81240-2|6
ATLANTIC|81240-2|7
ATLANTIC|81240-2|8
ATLANTIC|81240-2|9
ATLANTIC|81240-2|10
ATLANTIC|81240-2|11
A&M|6405-2|1
A&M|6405-2|2
A&M|6405-2|3
A&M|6405-2|5
A&M|6405-2|4
A&M|6405-2|7
A&M|6405-2|8
A&M|6405-2|6
A&M|6405-2|9
EMI|48057-2|1
EMI|48057-2|2
EMI|48057-2|3
REPRISE|25719-2|3
EMI|48057-2|6
EMI|48057-2|7
EMI|48057-2|8
EMI|48057-2|9
EMI|48057-2|10
REPRISE|25719-2|1
REPRISE|25719-2|2
REPRISE|25719-2|4
REPRISE|25719-2|5
REPRISE|25719-2|6
REPRISE|25719-2|7
REPRISE|25719-2|9
REPRISE|25719-2|10
EMI|48057-2|4
CBS|CK38837|1
CBS|CK38837|7
EOF

#
# try a non-existent database
#
if ./dbunload -f test_12.out no_such_database no_such_file 2> test_12.err; then
	exit 1
fi
cmp -s - test_12.err << \EOF
dbunload: can't open database no_such_database: IDBM File not found
EOF
#
# try a non-existent file
#
if ./dbunload -f test_12.out music no_such_file 2> test_12.err; then
	exit 1
fi
cmp -s - test_12.err << \EOF
dbunload: can't open file no_such_file: IDBM File does not exist
EOF
#
# now we try a non-existent index
#
if ./dbunload -f test_12.out music songs no_such_index 2> test_12.err; then
	exit 1
fi
cmp -s - test_12.err << \EOF
dbunload: can't open index no_such_index: IDBM Field does not exist
EOF
#
# try mentioning the same field twice in a view
#
if ./dbunload -f test_12.out -v sng_title,sng_title music songs 2> test_12.err
then
	exit 1
fi
cmp -s - test_12.err << \EOF
dbunload: can't create file view: IDBM Invalid argument
EOF
#
# and a non-existent field in a view
#
if ./dbunload -v no_such_field music songs 2> test_12.err; then
	exit 1
fi
cmp -s - test_12.err << \EOF
dbunload: unknown field no_such_field: IDBM Field does not exist
EOF
#
# if the shell supports the ulimit command, check out write errors
#
if (ulimit) > /dev/null 2>&1 ; then
	(
	ulimit 1
	if ./dbunload -f test_12.out music songs 2> test_12.err; then
		exit 1
	fi
	cmp -s - test_12.err << EOF
dbunload: write failed
EOF
	if ./dbunload -b -f test_12.out music songs 2> test_12.err; then
		exit 1
	fi
	cmp -s - test_12.err << EOF
dbunload: write failed
EOF
	)
fi
#
# Everything is OK
#
RC=0
